/*
 * Copyright 2013 <James.Bottomley@HansenPartnership.com>
 *
 * see COPYING file
 */
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>

#include <efi.h>

#include <kernel_efivars.h>
#include <guid.h>
#include <sha256.h>
#include "efiauthenticated.h"

/*****************************************************************************/

#define MAX_NAME    128

#define IOCTL_OEMHLP                0x0080

#define OEMHLP_GETFBDATA            0x0012
#define OEMHLP_GETVARIABLE          0x0017
#define OEMHLP_SETVARIABLE          0x0018

#define FILE_NORMAL                 0x0000
#define OPEN_ACTION_OPEN_IF_EXISTS  0x0001
#define OPEN_SHARE_DENYNONE         0x0040

uint32_t _System  DosDevIOCtl(uint32_t  hDevice,
                              uint32_t  category,
                              uint32_t  function,
                              void     *pParams,
                              uint32_t  cbParmLenMax,
                              uint32_t *pcbParmLen,
                              void     *pData,
                              uint32_t  cbDataLenMax,
                              uint32_t *pcbDataLen);

uint32_t _System  DosOpen(char     *pszFileName,
                          uint32_t *pHf,
                          uint32_t *pulAction,
                          uint32_t  cbFile,
                          uint32_t  ulAttribute,
                          uint32_t  fsOpenFlags,
                          uint32_t  fsOpenMode,
                          void     *peaop2);

uint32_t _System  DosClose(uint32_t hf);


/* OEMHLP_GETFBDATA packet */
typedef struct {
   uint16_t             Size;  // size of this structure (in/out)
   uint8_t             Flags;  // OEMHLP_FBF_*
   uint8_t            BitsPP;  // bits per pixel (should be 32)
   uint32_t          VMemLin;  // video memory linear address (in shared arena)
   uint32_t         VMemPhys;  // video memory physical address
   uint32_t            Pitch;  // scan line length
   uint16_t            SizeX,  // screen width
                       SizeY;  // screen height
   uint32_t          RedMask;
   uint32_t        GreenMask;
   uint32_t         BlueMask;
} OEMHLP_FBDATA;


/* OEMHLP_[G/S]ETVARIABLE packet */
typedef struct {
   uint64_t        ErrorCode;  // error code
   uint64_t          VarSize;  // variable data length
   uint32_t            Attrs;  // attributes
   uint8_t          Guid[16];  // vendor guid
   uint16_t          Name[1];  // variable name
} OEMHLP_VARDATA;


static uint32_t handle = 0;


int
os2_open_efivars(void)
{
    uint32_t        rc;
    uint32_t        u;
    OEMHLP_FBDATA   fb;

    rc = DosOpen("\\DEV\\OEMHLP$", &handle, &u, 0, FILE_NORMAL,
                 OPEN_ACTION_OPEN_IF_EXISTS, OPEN_SHARE_DENYNONE, NULL);
    if (rc) {
        fprintf(stderr, "Unable to open \\DEV\\OEMHLP$ - rc= %d\n", rc);
        return 0;
    }

    u = sizeof(OEMHLP_FBDATA);
    memset(&fb, 0, u);
    fb.Size = u;
    if (DosDevIOCtl(handle, IOCTL_OEMHLP, OEMHLP_GETFBDATA,
                    0, 0, 0, &fb, u, &u)) {
        fprintf(stderr, "This program is only usable on UEFI systems\n");
        DosClose(handle);
        return 0;
    }

    return handle;
}

void
os2_close_efivars(void)
{
    DosClose(handle);
    handle = 0;
    return;
}

int
get_variable(const char *name, EFI_GUID *guid, uint32_t *attributes,
	     uint32_t *out_len, void *out_buf)

{
    uint32_t ctr;
    uint32_t rc;
    uint32_t cbParm;
    uint8_t  parmBuf[sizeof(OEMHLP_VARDATA) + MAX_NAME * 2];
    OEMHLP_VARDATA *var = (OEMHLP_VARDATA*)parmBuf;

    memset(parmBuf, 0, sizeof(parmBuf));
    memcpy(var->Guid, (char*)guid, sizeof(EFI_GUID));
    cbParm = sizeof(OEMHLP_VARDATA);

    // quick & dirty conversion of ASCII to UTF16
    // (the terminating null is already included in cbParm)
    for (ctr = 0; ctr <= MAX_NAME && name[ctr]; ctr++) {
        var->Name[ctr] = (unsigned short)name[ctr];
        cbParm += 2;
    }

    if (!out_buf) {

        // get the size of the variable
        rc = DosDevIOCtl(handle, IOCTL_OEMHLP, OEMHLP_GETVARIABLE,
                         parmBuf, cbParm, &cbParm,
                         0, 0, 0);

        if (rc || (var->ErrorCode && var->ErrorCode != EFI_BUFFER_TOO_SMALL)) {

            if (var->ErrorCode != EFI_NOT_FOUND)
                fprintf(stderr, "OEMHLP_GETVARIABLE for size - rc= %d  errcode= %llx\n",
                        rc, var->ErrorCode);
            return -1;
        }

        *out_len = var->VarSize;
        if (attributes)
            *attributes = var->Attrs;

        return 0;
    }

    // get the variable's data
    rc = DosDevIOCtl(handle, IOCTL_OEMHLP, OEMHLP_GETVARIABLE,
                     parmBuf, cbParm, &cbParm,
                     out_buf, *out_len, out_len);

    if (rc || var->ErrorCode) {
        fprintf(stderr, "OEMHLP_GETVARIABLE - rc= %d  errcode= %llx\n",
                rc, var->ErrorCode);
        return -1;
    }

    if (*out_len != var->VarSize) {
        fprintf(stderr, "OEMHLP_GETVARIABLE - out_len= %d  VarSize= %lld\n",
                *out_len, var->VarSize);
        return -1;
    }

    if (attributes)
       *attributes = var->Attrs;

    return 0;
}

int
set_variable(const char *name, EFI_GUID *guid,
             uint32_t attributes, uint32_t out_len, void *out_buf)

{
    uint32_t ctr;
    uint32_t rc;
    uint32_t cbParm;
    uint8_t  parmBuf[sizeof(OEMHLP_VARDATA) + MAX_NAME * 2];
    OEMHLP_VARDATA *var = (OEMHLP_VARDATA*)parmBuf;

    memset(parmBuf, 0, sizeof(parmBuf));

    var->VarSize = out_len;
    var->Attrs = attributes;
    memcpy(var->Guid, (char*)guid, sizeof(EFI_GUID));
    cbParm = sizeof(OEMHLP_VARDATA);

    // quick & dirty conversion of ASCII to UTF16
    // (the terminating null is already included in cbParm)
    for (ctr = 0; ctr <= MAX_NAME && name[ctr]; ctr++) {
        var->Name[ctr] = (unsigned short)name[ctr];
        cbParm += 2;
    }

    // set the variable's data
    rc = DosDevIOCtl(handle, IOCTL_OEMHLP, OEMHLP_SETVARIABLE,
                     parmBuf, cbParm, &cbParm,
                     out_buf, out_len, &out_len);

    if (rc || var->ErrorCode) {
        fprintf(stderr, "OEMHLP_SETVARIABLE - rc= %d  errcode= %llx\n",
                rc, var->ErrorCode);
        return -1;
    }

    return 0;
}

/*****************************************************************************/

#if 0

static char *kernel_efi_path = NULL;

void
kernel_variable_init(void)
{
	char fname[] = "/tmp/efi.XXXXXX";
	char cmdline[256];
	int fd, ret;
	struct stat st;
	char *buf;

	if (kernel_efi_path)
		return;
	mktemp(fname);
	snprintf(cmdline, sizeof(cmdline), "mount -l > %s", fname);
	ret = system(cmdline);
	if (WEXITSTATUS(ret) != 0)
		/* hopefully stderr said what was wrong */
		exit(1);
	fd = open(fname, O_RDONLY);
	unlink(fname);
	if (fd < 0) {
		fprintf(stderr, "Failed to open output of %s\n", cmdline);
		exit(1);
	}
	if (fstat(fd, &st) < 0) {
		perror("stat failed");
		exit(1);
	}
	if (st.st_size == 0) {
		fprintf(stderr, "No efivarfs filesystem is mounted\n");
		exit(1);
	}
	buf = malloc(st.st_size);
	read(fd, buf, st.st_size);
	close(fd);

	char *ptr = buf;
	char path[512], type[512];
	while (ptr < buf + st.st_size) {
		int count;

		sscanf(ptr, "%*s on %s type %s %*[^\n]\n%n", path, type, &count);
		ptr += count;
		if (strcmp(type, "efivarfs") == 0)
			break;
	}
	if (strcmp(type, "efivarfs") != 0) {
		fprintf(stderr, "No efivarfs filesystem is mounted\n");
		exit(1);
	}
	kernel_efi_path = malloc(strlen(path) + 1);
	strcpy(kernel_efi_path, path);
}

int
get_variable(const char *var, EFI_GUID *guid, uint32_t *attributes,
	     uint32_t *size, void *buf)
{
	if (!kernel_efi_path)
		return -EINVAL;

	int varfs_len = strlen(var) + 48 + strlen(kernel_efi_path);
	char *varfs = malloc(varfs_len);
	uint32_t attr;
	int fd;
	struct stat st;

	snprintf(varfs, varfs_len, "%s/%s-%s", kernel_efi_path,
		 var, guid_to_str(guid));
	fd = open(varfs, O_RDONLY);
	free(varfs);
	if (fd < 0)
		return errno;
	
	if (fstat(fd, &st) < 0)
		return errno;
	if (size)
		*size = st.st_size - sizeof(attr);

	read(fd, &attr, sizeof(attr));

	if (attributes)
		*attributes = attr;

	if (buf)
		read(fd, buf, st.st_size - sizeof(attr));

	close(fd);

	return 0;
}

int
set_variable(const char *var, EFI_GUID *guid, uint32_t attributes,
	     uint32_t size, void *buf)
{
	if (!kernel_efi_path)
		return -EINVAL;

	int varfs_len = strlen(var) + 48 + strlen(kernel_efi_path);
	char *varfs = malloc(varfs_len),
		*newbuf = malloc(size + sizeof(attributes));
	int fd;

	snprintf(varfs, varfs_len, "%s/%s-%s", kernel_efi_path,
		 var, guid_to_str(guid));
	fd = open(varfs, O_RDWR|O_CREAT|O_TRUNC, 0644);
	free(varfs);
	if (fd < 0)
		return errno;
	memcpy(newbuf, &attributes, sizeof(attributes));
	memcpy(newbuf + sizeof(attributes), buf, size);
	
	if (write(fd, newbuf, size + sizeof(attributes)) != size + sizeof(attributes))
		return errno;

	close(fd);

	return 0;
}

#endif

/*****************************************************************************/

int
get_variable_alloc(const char *var, EFI_GUID *guid, uint32_t *attributes,
		   uint32_t *size, uint8_t **buf)
{
	uint32_t len;
	int ret = get_variable(var, guid, NULL, &len, NULL);
	if (ret)
		return ret;

    *size = len;

	*buf = malloc(len);
	if (!buf)
		return -ENOMEM;

	return get_variable(var, guid, attributes, size, *buf);
}

int
variable_is_setupmode(void)
{
	uint8_t setup_mode = 0;
    uint32_t len = 1;

	get_variable("SetupMode", &GV_GUID, NULL, &len, &setup_mode);

	return setup_mode;
}

int
variable_is_secureboot(void)
{
	uint8_t secure_boot = 0;
    uint32_t len = 1;

	get_variable("SecureBoot", &GV_GUID, NULL, &len, &secure_boot);

	return secure_boot;
}


int
set_variable_esl(const char *var, EFI_GUID *guid, uint32_t attributes,
		 uint32_t size, void *buf)
{
//	if (!kernel_efi_path)
//		return -EINVAL;

	int newsize = size + OFFSET_OF(EFI_VARIABLE_AUTHENTICATION_2, AuthInfo) + OFFSET_OF(WIN_CERTIFICATE_UEFI_GUID, CertData);
	char *newdata = malloc(newsize);
	EFI_VARIABLE_AUTHENTICATION_2 *DescriptorData;
	EFI_TIME *Time;
	struct tm tm;
	time_t t;

	time(&t);

	memset(newdata, '\0', newsize);
	memcpy(newdata + newsize - size, buf, size);
	DescriptorData = (EFI_VARIABLE_AUTHENTICATION_2 *)newdata;
	Time = &DescriptorData->TimeStamp;
	gmtime_r(&t, &tm);
	/* FIXME: currently timestamp is one year into future because of
	 * the way we set up the secure environment  */
	Time->Year = tm.tm_year + 1900 + 1;
	/* EFI_TIME Month is 1-12; Unix tm_mon is 0-11 */
	Time->Month = tm.tm_mon + 1;
	Time->Day = tm.tm_mday;
	Time->Hour = tm.tm_hour;
	Time->Minute = tm.tm_min;
	Time->Second = tm.tm_sec;
	DescriptorData->AuthInfo.Hdr.dwLength  = OFFSET_OF (WIN_CERTIFICATE_UEFI_GUID, CertData);
	DescriptorData->AuthInfo.Hdr.wRevision = 0x0200;
	DescriptorData->AuthInfo.Hdr.wCertificateType = WIN_CERT_TYPE_EFI_GUID;
	DescriptorData->AuthInfo.CertType =  EFI_CERT_TYPE_PKCS7_GUID;

	int ret = set_variable(var, guid, attributes, newsize, newdata);
	free(newdata);
	return ret;
}

uint8_t *
hash_to_esl(EFI_GUID *owner, int *len,
	    uint8_t hash[SHA256_DIGEST_SIZE])
{
	const int siglen = sizeof(EFI_SIGNATURE_LIST) + sizeof(EFI_SIGNATURE_DATA) - 1 + SHA256_DIGEST_SIZE;
	uint8_t *sig = malloc(siglen);
	EFI_SIGNATURE_LIST *l = (void *)sig;
	EFI_SIGNATURE_DATA *d = (void *)sig + sizeof(EFI_SIGNATURE_LIST);

	if (len)
		*len = siglen;

	memset(sig, 0, siglen);
	l->SignatureType = EFI_CERT_SHA256_GUID;
	l->SignatureListSize = siglen;
	l->SignatureSize = 16 +32; /* UEFI defined */
	memcpy(&d->SignatureData, hash, SHA256_DIGEST_SIZE);
	d->SignatureOwner = *owner;

	return sig;
}

int
set_variable_hash(const char *var, EFI_GUID *owner, uint32_t attributes,
		  uint8_t hash[SHA256_DIGEST_SIZE])
{
	int len;
	uint8_t *sig = hash_to_esl(&MOK_OWNER, &len, hash);

	int ret = set_variable_esl(var, owner, attributes, len, sig);
	free(sig);
	return ret;
}
